@@ -1,3 +1,40 @@
+;;; mockery.scm
+;;;
+;;;  s7.html says that this file implements "openlets masquerading as various data
+;;;  types".  The original idea was to test the generic function support in s7 by
+;;;  creating objects that act as if they were (say) vectors or strings.  Each such
+;;;  object is a let (openlet) inheriting its methods (functions like vector-length)
+;;;  from the outer let (mock-vector-class etc), and within the object itself there
+;;;  are a 'value field and a 'mock-type field.  Each class (a let) has an associated
+;;;  make function (make-mock-vector or mock-vector etc). The class functions give
+;;;  you an object that acts everywhere in scheme like (say) a vector, but you can
+;;;  specialize it by either adding more functions, or changing the ones given, or
+;;;  adding other fields to carry around arbitrary information with the object.
+;;;  
+;;;  To make a mock-vector call (*mock-vector* 'make-mock-vector) or 
+;;;  (*mock-vector* 'mock-vector):
+;;;  
+;;;    ((*mock-vector* 'mock-vector) 1 2 3) -> #(1 2 3)
+;;;  
+;;;  This prints as #(1 2 3) but it's actually one of these mock objects:
+;;;  
+;;;    (let? ((*mock-vector* 'mock-vector) 1 2 3)) -> #t
+;;;  
+;;;  There are examples scattered around this file, and a lot more
+;;;  in s7test.scm.
+;;;
+;;; currently it is possible to (set! ((*mock-vector* 'mock-vector-class) 'write) hash-table-set!) [or (set! (v 'write) hash-table-set!) I think]
+;;;   should we call the s7.html vars-immutable (and immutable!) on this internal let?
+;;; to create an immutable mock datum, call immutable! on its value:
+;;;   (define-constant L (let ((tmp ((*mock-pair* 'mock-pair) 1 2))) (immutable! (tmp 'value)) tmp))
+;;;
+;;; also we omit a variety of built-in functions that can work on the underlying value.
+;;;   given (define v #(0 1)) and (define mv ((*mock-vector* 'mock-vector) 0 1))
+;;;     (arity v): '(1 . 536870912), (arity mv): '(1 . 1)
+;;;     (signature v): '(#t vector? . #1=(integer? . #1#)), (signature mv): '(#t let? . #1=(symbol? . #1#))
+;;;     object->let but this needs to be omitted for debugging
+;;;     and many where the error message is different (should pretty-print know about these?)
+
 (provide 'mockery.scm)
 
 ;;; the exported mock data classes
@@ -29,21 +66,15 @@
   (define (with-mock-wrapper func)
     (lambda (obj)
       (cond ((mock? obj)
-	     (dynamic-wind
-		 coverlets
-		 (lambda ()
-		   (func (obj 'value)))
-		 openlets))
+	     (let-temporarily (((*s7* 'openlets) #f))
+	       (func (let-ref obj 'value))))
 
 	    ((not (openlet? obj))
 	     (func obj))
 
 	    ((procedure? obj) ; TODO: and c-pointer? c-object?
-	     (dynamic-wind
-		 coverlets
-		 (lambda ()
-		   (func obj))
-		 openlets))
+	     (let-temporarily (((*s7* 'openlets) #f))
+	       (func obj)))
 
 	    (else
 	     (let ((func-name (string->symbol (object->string func))))
@@ -53,12 +84,16 @@
 
   (define (with-mock-wrapper* func)
     (lambda args
+      ;(format *stderr* "with-mock-wrapper ~S~%" args)
       (let ((unknown-openlets #f)
-	    (new-args ()))
+	    (new-args ())
+	    (got-mock #f))
 	(for-each (lambda (arg) ; not map here because (values) should not be ignored: (+ (mock-number 4/3) (values))
 		    (set! new-args
 			  (cons (if (mock? arg)
-				    (arg 'value)
+				    (begin
+				      (set! got-mock #t)
+				      (arg 'value))
 				    (begin
 				      (if (and (openlet? arg)
 					       (not (procedure? arg))
@@ -68,13 +103,27 @@
 				      arg))
 				new-args)))
 		  args)
-	(if unknown-openlets
-	    (apply func (reverse new-args))
-	    (dynamic-wind coverlets (lambda () (apply func (reverse new-args))) openlets)))))
+	(if (and unknown-openlets got-mock)
+	    (apply func (reverse! new-args))
+	    (let-temporarily (((*s7* 'openlets) #f)) 
+	      (apply func (reverse! new-args)))))))
+
+
+  (define (append-wrapper . args)
+    ;;  append is odd -- if 0 args -> (), 1 arg -> arg, 2 args and first is (), return arg2(etc), else (apply #_append args-via-values)?
+    ;;    we can get here with 0 args via: (with-let (*mock-vector* 'mock-vector-class) (append)) !
+    (if (null? args)
+	()
+	(if (null? (cdr args)) 
+	    (car args) ; presumably a mock object
+	    (if (and (null? (car args))  ; etc...
+		     (null? (cddr args)))
+		(cadr args)
+		(apply (with-mock-wrapper* #_append) args)))))
 
   ;; one tricky thing here is that a mock object can be the let of with-let: (with-let (mock-port ...) ...)
-  ;;   so a mock object's method can be called even when no argument is a mock object.  Even trickier, the
-  ;;   mock object can be a closure's containing (open)let: (display (openlet (with-let (mock-c-pointer 0) (lambda () 1))))
+  ;;   so a mock object's method can be called even when no argument is a mock object.  Even trickier
+  ;;   (display (openlet (with-let (mock-c-pointer 0) (lambda () 1))))
   ;; --------------------------------------------------------------------------------
 
   (set! *mock-vector*
@@ -85,13 +134,13 @@
 						 (error 'wrong-type-arg "stray mock-vector? ~S" i))
 					     (#_vector-set! (->value obj) i val))
 		       
-		       'vector-set!        (lambda (obj i val) ((obj 'local-set!) obj i val) val)
+		       'vector-set!        (lambda (obj i val) ((let-ref obj 'local-set!) obj i val) val)
 		       
 		       'let-set-fallback   (lambda (obj i val) 
 					     (if (and (integer? i)
 						      (defined? 'value obj))
 						 (begin
-						   ((obj 'local-set!) obj i val) 
+						   ((let-ref obj 'local-set!) obj i val) 
 						   val)
 						 (error 'out-of-range "unknown field: ~S" i)))
 		       
@@ -117,7 +166,9 @@
 		       'format             (with-mock-wrapper* #_format)
 		       'write              (with-mock-wrapper* #_write)
 		       'display            (with-mock-wrapper* #_display)
+		       'vector-dimension   (with-mock-wrapper #_vector-dimension)
 		       'vector-dimensions  (with-mock-wrapper #_vector-dimensions)
+		       'vector-rank        (with-mock-wrapper #_vector-rank)
 		       'fill!              (with-mock-wrapper* #_fill!)
 		       'vector-fill!       (with-mock-wrapper* #_vector-fill!)
 		       'vector->list       (with-mock-wrapper* #_vector->list)
@@ -128,8 +179,16 @@
 		       'byte-vector?       (with-mock-wrapper #_byte-vector?)
 		       'float-vector?      (with-mock-wrapper #_float-vector?)
 		       'length             (with-mock-wrapper #_length)
-		       'vector-append      (with-mock-wrapper* #_vector-append)
-		       'append             (with-mock-wrapper* #_append)
+		       'vector-append      (if (provided? 'pure-s7)
+					       (lambda args
+						 (if (null? args) 
+						     #() 
+						     (if (vector? (car args))
+							 (apply append args)
+							 (error 'wrong-type-arg "vector-append arguments should be vectors: ~A" args))))
+					       (with-mock-wrapper* #_vector-append))
+		       'append             append-wrapper ;(with-mock-vector #_append)
+		       'vector-typer       (with-mock-wrapper #_vector-typer)
 		       'class-name         '*mock-vector*)))
 	  
 	  (define (make-mock-vector len . rest)
@@ -151,6 +210,7 @@
 	  
 	  (curlet)))
 
+;;; perhaps mock-float-vector etc?
 
 #|
   ;; vector that grows to accommodate vector-set!
@@ -172,6 +232,20 @@
 	 'let-ref-fallback local-ref
 	 'vector-set! local-set!
 	 'let-set-fallback local-set!))))
+
+  ;; vector with data (mimicking Clojure):
+  (define (meta-vector v data)
+    (openlet
+     (sublet (*mock-vector* 'mock-vector-class)
+       'value v
+       'mock-type 'mock-vector?
+       'meta-data data)))
+  
+  (define v (meta-vector #(0 1 2) "hiho"))
+  (v 1): 1
+  (v 'meta-data): "hiho"
+  v: #(0 1 2)
+  (vector? v): #t
 |#
 
 
@@ -211,7 +285,9 @@
 		       'copy               (with-mock-wrapper* #_copy)
 		       'hash-table?        (with-mock-wrapper #_hash-table?)
 		       'length             (with-mock-wrapper #_length)
-		       'append             (with-mock-wrapper* #_append)
+		       'append             append-wrapper ;(with-mock-wrapper* #_append)
+		       'hash-table-key-typer (with-mock-wrapper #_hash-table-key-typer)
+		       'hash-table-value-typer (with-mock-wrapper #_hash-table-value-typer)
 		       'class-name         '*mock-hash-table*)))
 	  
 	  (define (make-mock-hash-table . rest)
@@ -253,7 +329,7 @@
 			  (#_hash-table-ref (obj 'value) key)))))
   
   (define (hash-table-key? obj key) 
-    ((obj 'hash-table-key?) obj key))
+    ((let-ref obj 'hash-table-key?) obj key))
   
   (define* (make-gloomy-hash-table (len 511) not-a-key)
     (let ((ht (gloomy-hash-table)))
@@ -290,10 +366,10 @@
 						     (if (string? str)
 							 (length str)
 							 (error 'wrong-type-arg "string-length argument should be a string: ~A" str)))
-						   (with-mock-wrapper #_string-length))
+						   (with-mock-wrapper #_length))
 
 		       'string-append          (with-mock-wrapper* #_string-append)
-		       'string-copy            (with-mock-wrapper #_copy)
+		       'string-copy            (with-mock-wrapper* #_string-copy)
 		       
 		       'string=?               (with-mock-wrapper* #_string=?)
 		       'string<?               (with-mock-wrapper* #_string<?)
@@ -304,7 +380,7 @@
 		       'string-downcase        (with-mock-wrapper #_string-downcase)
 		       'string-upcase          (with-mock-wrapper #_string-upcase)
 		       'string->symbol         (with-mock-wrapper #_string->symbol)
-		       'symbol                 (with-mock-wrapper #_symbol)
+		       'symbol                 (with-mock-wrapper* #_symbol)
 		       'string->keyword        (with-mock-wrapper #_string->keyword)
 		       'open-input-string      (with-mock-wrapper #_open-input-string)
 		       'directory?             (with-mock-wrapper #_directory?)
@@ -325,7 +401,7 @@
 		       'load                   (with-mock-wrapper* #_load)
 		       'eval-string            (with-mock-wrapper* #_eval-string)
 		       'string->list           (with-mock-wrapper* #_string->list)
-		       'bignum                 (with-mock-wrapper #_bignum)
+		       'bignum                 (with-mock-wrapper* #_bignum)
 		       'fill!                  (with-mock-wrapper* #_fill!)
 		       'write-string           (with-mock-wrapper* #_write-string)
 		       'copy                   (with-mock-wrapper* #_copy)
@@ -340,8 +416,8 @@
 		       'string-ci<=?           (with-mock-wrapper* #_string-ci<=?)
 		       'string-ci>=?           (with-mock-wrapper* #_string-ci>=?)
 		       'string?                (with-mock-wrapper #_string?)
-		       'length                 (with-mock-wrapper #_string-length)
-		       'append                 (with-mock-wrapper* #_append)
+		       'length                 (with-mock-wrapper (if (provided? 'pure-s7) #_length #_string-length))
+		       'append                 append-wrapper ;(with-mock-wrapper* #_append)
 		       'class-name             '*mock-string*)))
 	  
 	  (define* (make-mock-string len (init #\null))
@@ -415,6 +491,7 @@
 		       'char-ci>=?         (with-mock-wrapper* #_char-ci>=?)
 		       'string             (with-mock-wrapper* #_string)
 		       'string-fill!       (with-mock-wrapper* #_string-fill!)
+		       'fill!              (with-mock-wrapper* #_fill!)
 		       'object->string     (with-mock-wrapper* #_object->string)
 		       'format             (with-mock-wrapper* #_format)
 		       'write              (with-mock-wrapper* #_write)
@@ -437,7 +514,7 @@
 		  (sublet (*mock-char* 'mock-char-class)
 		    'value c
 		    'mock-type 'mock-char?)))
-		(error 'wrong-type-arg "mock-char arg ~S is not a char" c)))
+		(error 'wrong-type-arg "mock-char argument ~S is not a char" c)))
 	  
 	  (set! mock-char? (lambda (obj)
 			     (and (let? obj)
@@ -478,7 +555,7 @@
 		 'negative?        (with-mock-wrapper #_negative?)
 		 'infinite?        (with-mock-wrapper #_infinite?)
 		 'nan?             (with-mock-wrapper #_nan?)
-		 'append           (with-mock-wrapper* #_append)
+		 ;'append          append-wrapper ; (with-mock-wrapper* #_append) ;?? (append ... 3 ...) is an error
 		 'magnitude        (with-mock-wrapper #_magnitude)
 		 'angle            (with-mock-wrapper #_angle)
 		 'rationalize      (with-mock-wrapper* #_rationalize)
@@ -513,6 +590,8 @@
 		 'real?            (with-mock-wrapper #_real?)
 		 'complex?         (with-mock-wrapper #_complex?)
 		 'rational?        (with-mock-wrapper #_rational?)
+		 'byte?            (with-mock-wrapper #_byte?)
+		 'float?           (with-mock-wrapper #_float?)
 		 'exact?           (with-mock-wrapper #_exact?)
 		 'inexact?         (with-mock-wrapper #_inexact?)
 		 'lognot           (with-mock-wrapper #_lognot)
@@ -534,11 +613,13 @@
 		 '<=               (with-mock-wrapper* #_<=)
 		 '>=               (with-mock-wrapper* #_>=)
 		 
-		 'make-polar       (if (provided? 'pure-s7)
-				       (lambda (mag ang) (#_complex (* mag (cos ang)) (* mag (sin ang))))
-				       (lambda args (apply #_make-polar (map ->value args))))
+ 		 'make-polar       (if (provided? 'pure-s7)
+ 				       (lambda (mag ang) (#_complex (* mag (cos ang)) (* mag (sin ang))))
+ 				       (lambda (mag ang) (#_make-polar (->value mag) (->value arg))))
+
 		 'make-rectangular (with-mock-wrapper* #_complex)
 		 'complex          (with-mock-wrapper* #_complex)
+		 'bignum           (with-mock-wrapper* #_bignum)
 		 'random-state     (with-mock-wrapper* #_random-state)
 		 'ash              (with-mock-wrapper* #_ash)
 		 'logbit?          (with-mock-wrapper* #_logbit?)
@@ -563,6 +644,7 @@
 		 'string->list     (with-mock-wrapper* #_string->list)
 		 'substring        (with-mock-wrapper* #_substring)
 		 'vector-fill!     (with-mock-wrapper* #_vector-fill!)
+		 'fill!            (with-mock-wrapper* #_fill!)
 		 'make-string      (with-mock-wrapper* #_make-string)
 		 'string-ref       (with-mock-wrapper* #_string-ref)
 		 'string-set!      (with-mock-wrapper* #_string-set!)
@@ -584,7 +666,7 @@
 		 'subvector        (with-mock-wrapper* #_subvector)
 		 'read-string      (with-mock-wrapper* #_read-string)
 		 'length           (with-mock-wrapper #_length)
-		 'number?          (with-mock-wrapper #_number?)
+		 'c-pointer        (with-mock-wrapper #_c-pointer)
 		 'class-name       '*mock-number*)))
 	  
 	  (define (mock-number x)
@@ -622,7 +704,6 @@
   ;; from Wikipedia:
   ;; x + y =	[a+c, b+d]
   ;; x - y =	[a-d, b-c]
-  ;; x  y =	[min(ac, ad, bc, bd), max(ac, ad, bc, bd)]
   ;; x / y =	[min(a/c, a/d, b/c, b/d), max(a/c, a/d, b/c, b/d)]
   
   (define *interval*
@@ -783,7 +864,7 @@
 								(#_list-ref (obj 'value) ind))))
 						     (openlet obj) 
 						     val)
-						   (error "let-ref mock-pair index is not an integer: ~S" ind))))
+						   (error 'wrong-type-arg "let-ref mock-pair index is not an integer: ~S" ind))))
 		       'let-set-fallback (lambda (obj ind val) 
 					   (if (eq? ind 'value)
 					       #<undefined>
@@ -793,7 +874,7 @@
 								(#_list-set! (obj 'value) ind val))))
 						     (openlet obj)
 						     val)
-						   (error "let-set! mock-pair index is not an integer: ~S" ind))))
+						   (error 'wrong-type-arg "let-set! mock-pair index is not an integer: ~S" ind))))
 		       
 		       'reverse!         (lambda (obj) 
 					   (if (mock-pair? obj)
@@ -815,13 +896,13 @@
 		       'list-set!        (with-mock-wrapper* #_list-set!)
 		       'pair?            (with-mock-wrapper #_pair?)
 		       'length           (with-mock-wrapper #_length)
-		       'append           (with-mock-wrapper* #_append)
+		       'append           append-wrapper ;(with-mock-wrapper* #_append)
 		       'class-name       '*mock-pair*)))
 	  
 	  (define (mock-pair . args)
 	    (openlet
 	     (sublet (*mock-pair* 'mock-pair-class)
-	       'value (copy args)
+	       'value (copy args) ; i.e. like list not cons
 	       'mock-type 'mock-pair?)))
 	  
 	  (set! mock-pair? (lambda (obj)
@@ -870,14 +951,15 @@
 	       (mock-symbol-class
 		(inlet 'equivalent?           (with-mock-wrapper* #_equivalent?)
 		       'gensym?               (with-mock-wrapper #_gensym?)
-		       'append                (with-mock-wrapper* #_append)
+		       ;'append               (with-mock-wrapper* #_append) ;? (append ... 'a ...) is an error
+		       'fill!                 (with-mock-wrapper* #_fill!)
 		       'symbol->string        (with-mock-wrapper #_symbol->string)
 		       'symbol->value         (with-mock-wrapper* #_symbol->value)
 		       'symbol->dynamic-value (with-mock-wrapper #_symbol->dynamic-value)
-		       'setter                (with-mock-wrapper #_setter)
+		       'setter                (with-mock-wrapper* #_setter)
 		       'provided?             (with-mock-wrapper #_provided?)
 		       'provide               (with-mock-wrapper #_provide)
-		       'defined?              (with-mock-wrapper #_defined?)
+		       'defined?              (with-mock-wrapper* #_defined?)
 		       'symbol->keyword       (with-mock-wrapper #_symbol->keyword)
 		       'keyword?              (with-mock-wrapper #_keyword?)
 		       'keyword->symbol       (with-mock-wrapper #_keyword->symbol)
@@ -921,6 +1003,8 @@
 		       'format          (with-mock-wrapper* #_format)
 		       'write           (with-mock-wrapper* #_write)
 		       'display         (with-mock-wrapper* #_display)
+		       'fill!           (with-mock-wrapper* #_fill!)
+		       'class-name      '*mock-c-pointer*
 		       )))
 	  
 	  (define* (mock-c-pointer (int 0) type info weak1 weak2)
@@ -951,13 +1035,14 @@
 		       'format             (with-mock-wrapper* #_format)
 		       'write              (with-mock-wrapper* #_write)
 		       'display            (with-mock-wrapper* #_display)
+		       'class-name         '*mock-random-state*
 		       )))
 	  
 	  (define* (mock-random-state seed (carry 1675393560))
 	    (immutable!
 	     (openlet 
 	      (sublet (*mock-random-state* 'mock-random-state-class)
-		'value (#_random-state seed carry)
+		'value (if (provided? 'gmp) (#_random-state seed) (#_random-state seed carry))
 		'mock-type 'mock-random-state?))))
 	  
 	  (set! mock-random-state? 
@@ -982,6 +1067,7 @@
 		       'format            (with-mock-wrapper* #_format)
 		       'write             (with-mock-wrapper* #_write)
 		       'display           (with-mock-wrapper* #_display)
+		       'class-name        '*mock-iterator*
 		       )))
 	  
 	  (define (make-mock-iterator . args)
@@ -1009,7 +1095,7 @@
 		       'output-port?        (with-mock-wrapper #_output-port?)
 		       'port-closed?        (with-mock-wrapper #_port-closed?)
 		       'equivalent?         (with-mock-wrapper* #_equivalent?)
-		       'append              (with-mock-wrapper* #_append)
+		       ;'append              (with-mock-wrapper* #_append) ; ?? (append (open-input-string "asdf") ...) is an error
 		       'set-current-output-port (with-mock-wrapper #_set-current-output-port)
 		       'set-current-input-port  (with-mock-wrapper #_set-current-input-port)
 		       'set-current-error-port  (with-mock-wrapper #_set-current-error-port)
@@ -1102,4 +1188,4 @@
   ((p 'owner) p) -> "bil"
 |#
 
-  ) ; end of outer let
+  #f) ; end of outer let
